/* Includes {{{*/
#include "common.h"
/*}}}*/

/* Constants *that may need to be modified* {{{*/
// TODO Include extra time to send nacks
// TODO FIND FIRST PACKET GOOD AND FILL IT IN

/*
 * Constants for determining correctness of 
 * retransmission headers.
 */

/* 802.11 */
u_char SRC[6];
u_char BSSID[6];
u_char DST[6];

/* LLC */
u_char LLC_ORG[3];
u_char LLC_TYPE[2];

/* IP */
u_char IP_SRC[4];
u_char IP_DST[4];
uint16_t IP_LEN;

/* UDP */
uint16_t UDP_LEN;
uint16_t UDP_SRC;
uint16_t UDP_DST;

/* Bad blocks */
int (*good_blocks)(u_char* packet, int len) = num_good_blocks_iperf;
int (*good_blocks_in_first_n_blocks)(u_char* packet, int len, int blocks) = num_good_blocks_iperf_in_first_n_blocks;
int (*good_contig_blocks)(u_char* packet, int len) = num_good_contig_blocks_iperf;
/*}}}*/

struct simulator_state /*{{{*/
{
	uint64_t mactime; /* Current time of the simulated MAC */
	uint64_t mactime_actual; /* Current time of the actual MAC */
	uint32_t packets_total;
	uint32_t packets_nomactime;
	uint32_t packets;
	uint32_t packets_bad;
	uint32_t packets_orig;
	uint32_t packets_retran;
	uint32_t packets_retran_bad;
	uint32_t packets_retran_follows_bad;

	uint32_t bytes_total;
	uint32_t rates_total;
	uint32_t orig_rates_total;
	uint32_t blocks_total;
	uint32_t error_blocks_total;

	uint64_t time_saved_repair_worked;
	uint64_t time_saved_one_repair_worked;
	uint64_t time_saved_two_repair_worked;
	uint64_t time_saved_retransmit_less;

	//int retrydone;

	uint64_t prev_mactime;
	uint64_t prev_mactime_actual;
	uint16_t prev_seq;
	int prev_badblocks;

	int initialized_constants;
}; /*}}}*/

/* Function Headers {{{*/
void simulate_packet(struct pcap_pkthdr* pcap_header, u_char* packet, struct simulator_state* state);
void print_simulator_state(struct simulator_state* state);
int is_header_correct(u_char* packet, int len);
void print_mac(const u_char* mac);
inline long double bandwidth(uint32_t bytes, uint64_t time);
void die(char *text);
uint64_t packet_retransmission_time_us(int blocks, double bitrate);
uint64_t packet_transmission_time_us(int len, double bitrate);
void init_constants_from_packet(u_char* packet, int len);
/*}}}*/

int main(int argc, char** argv) /*{{{*/
{
	if (argc < 2)
	{
		printf("Usage: %s <pcap>\n", argv[0]);
		exit(0);
	}

	char pcap_error[PCAP_ERRBUF_SIZE];
	pcap_t* pcap = NULL;
	struct bpf_program bpf;
	u_char* packet = NULL;
	struct pcap_pkthdr pcap_header;
	struct simulator_state state;

	memset(&state, 0, sizeof(struct simulator_state));

	pcap = pcap_open_offline(argv[1], pcap_error);
	if (pcap == NULL)
	{
		printf("pcap failed: %s\n", pcap_error);
		return 2;
	}

	/* Filter out only the UDP packets */
	//pcap_compile(pcap, &bpf, PCAP_FILTER, 0, 0);
	//pcap_setfilter(pcap, &bpf);

	while ((packet = (u_char*)pcap_next(pcap, &pcap_header)) != NULL)
	{
		simulate_packet(&pcap_header, packet, &state);
	}

	print_simulator_state(&state);

	pcap_close(pcap);	

	return 0;
} /*}}}*/

void simulate_packet(struct pcap_pkthdr* pcap_header, u_char* packet, struct simulator_state* state) /*{{{*/
{
	/* Only look at packets that match the packet size we specified */
	if (pcap_header->caplen != PACKET_SIZE)
		return;	

	state->packets_total++;

	int rtaplen = radiotap_len(packet);
	int len = pcap_header->caplen;

	if (!radiotap_hasflag(packet, IEEE80211_RADIOTAP_TSFT))
	{
		state->packets_nomactime++;
		return;
	}

	/* 
	 * Assumption:
	 * For atheros (madwifi) monitor the packet start time is the end of the packet
	 */

	struct wlan_hdr* wlan = (struct wlan_hdr*)(packet + rtaplen);
	/* ! */
	double rate = radiotap_rate(packet);
	uint64_t mactime = radiotap_mactime(packet) - packet_transmission_time_us(len, rate);
	uint64_t endmactime = radiotap_mactime(packet);
	uint32_t fcs = ntohl(*(uint32_t*)(packet + len - sizeof(uint32_t)));
	/* Crc is computed over MAC header + payload (-4 for crc already in packet) */
	uint32_t calc_fcs = crc32_802(packet + rtaplen, len - rtaplen - 4);
	uint16_t seqnum = wlan_seqnum(wlan->seq);
	u_char retry = wlan_retry(wlan->fc);
	uint32_t badblocks = num_blocks(packet, len) - good_blocks(packet, len);

	//printf("Good blocks %d out of %d\n", good_blocks(packet, len), num_blocks(packet, len));

	state->blocks_total += num_blocks(packet, len);
	state->error_blocks_total += badblocks;
	state->rates_total += rate;

	/* 
	 *	Increment transmitted count,
	 *		if next seq and a good packet.
	 *	Assumptions:
	 *		1. At some point a packet will be transmitted correctly
	 *		2. No missed acks that could cause multiple good retransmissions for the same seqnum
	 */

	/* Good packet or good retry*/
	if (fcs == calc_fcs)
	{
		state->bytes_total += pcap_header->len;
	}

	/* 
	 *
	 * Compute simulated time of packet arrival
	 *
	 * Start with the time from the transmission time from the trace,
	 * 	then remove time as neccesary.
	 *
	 * Assumptions:
	 * 	1. If a retran is received correctly the ack will received by
	 * 		the sender (so we are eliminating retrans caused by missed ACKs)
	 * 	2. Sequence number in bad packets are good
	 *  3. NACKS are never lost.
	 */	

	/* Compute estimated time of packet arrival, some of this time may be removed */
	if (state->prev_mactime != 0)
		state->mactime += (endmactime - state->prev_mactime);
	else
		state->mactime = endmactime - mactime;

	if (fcs == calc_fcs)
	{

		/* Init constants used to check correctness of headers */
		if (!state->initialized_constants)
		{
			init_constants_from_packet(packet, len);
			state->initialized_constants = 1;
		}

		// Seqnum of good packets changes, do not erase time 
		if (state->prev_seq != seqnum)
			state->prev_badblocks = 0;

		/* Good retransmission {{{*/
		if (retry)
		{
			printf("%lld PACKET %d %lld GOOD_RETRY %f\n", state->mactime_actual, seqnum, 
					endmactime - mactime, rate);
			printf("%lld SIM-PACKET %d %lld GOOD_RETRY\n", state->mactime_actual, seqnum, 
					endmactime - mactime);

			state->packets_retran++;

			/* Log retran follows bad packet */
			if (state->prev_badblocks > 0)
				state->packets_retran_follows_bad++;

			/* Would this be a successful NACK? */
			if (state->prev_mactime != 0)
			{
				// Already done retry
				if (state->prev_badblocks < 0)
				{
					printf("%lld SIM-PACKET %d %lld SPACE\n", state->mactime_actual, seqnum, 
							endmactime - state->prev_mactime_actual);
					state->mactime -= (endmactime - state->prev_mactime_actual);
					state->time_saved_repair_worked  += (endmactime - state->prev_mactime_actual);
				}
				if (state->prev_badblocks > 0)
				{
					printf("%lld SIM-PACKET %d %lld NACK\n", state->mactime_actual, seqnum,
							packet_retransmission_time_us(state->prev_badblocks, rate));
					printf("%lld SIM-PACKET %d %lld SPACE\n", state->mactime_actual + 
							packet_retransmission_time_us(state->prev_badblocks, rate), 
							seqnum, 
							endmactime - mactime - 
							packet_retransmission_time_us(state->prev_badblocks, rate));

					state->time_saved_retransmit_less  += (endmactime - mactime) - 
						packet_retransmission_time_us(state->prev_badblocks, rate);

					assert(state->prev_badblocks * BLOCKSIZE <= len); /* could be tighter than this. */
					printf("Time to transmit packet %lld\n", endmactime - mactime);
					printf("Time to retransmit small packet %lld\n", packet_retransmission_time_us(state->prev_badblocks, rate));
					printf("Blocks to transmit %d blocks at rate %f\n", state->prev_badblocks, rate);

					/* Remove the extra time not spent sending the NACK */ 
					state->mactime -= (endmactime - mactime) - 
						packet_retransmission_time_us(state->prev_badblocks, rate);
					// Done the retry, assume nack gets back
					state->prev_badblocks = -1;
				}
			}
		} /*}}}*/

		/* Good transmission {{{*/
		else
		{
			state->orig_rates_total += rate;
			state->packets_orig++;
			printf("%lld PACKET %d %lld GOOD_DATA %f\n", state->mactime_actual, seqnum, 
					endmactime - mactime, rate);
			printf("%lld SIM-PACKET %d %lld GOOD_DATA\n", state->mactime_actual, seqnum, 
					endmactime - mactime);

			state->packets++;
		} /*}}}*/

		state->prev_seq = seqnum;
	} 

	else
	{
		/* Bad retransmission {{{*/
		if (retry)
		{
			printf("%lld PACKET %d %lld BAD_RETRY %f %d %d %d\n", state->mactime_actual, seqnum, 
					endmactime - mactime, rate, is_header_correct(packet, len), good_contig_blocks(packet, len), num_blocks(packet, len));
			printf("%lld SIM-PACKET %d %lld BAD_RETRY %d %d %d\n", state->mactime_actual, seqnum, 
					endmactime - mactime, is_header_correct(packet, len), good_contig_blocks(packet, len), num_blocks(packet, len));
			print_good_blocks_iperf(packet, len, rate, state->mactime_actual);

			state->packets_retran_bad++;

			/* Log retran follows bad packet */
			if (state->prev_badblocks > 0)
				state->packets_retran_follows_bad++;

			/* Would this be a successful NACK? */
			/* Only successful nack if there are enough contig correct blocks and header is good */
			// do not care if header is correct for nack
			if (state->prev_badblocks < 0)
			{
				printf("%lld SIM-PACKET %d %lld SPACE \n", state->mactime_actual, seqnum, 
						endmactime - state->prev_mactime_actual);
				state->mactime -= (endmactime - state->prev_mactime_actual);
				state->time_saved_repair_worked += (endmactime - state->prev_mactime_actual);
			}
			if (state->prev_badblocks > 0)	
			{
				printf("%lld SIM-PACKET %d %lld NACK \n", state->mactime_actual, seqnum, 
						packet_retransmission_time_us(state->prev_badblocks, rate));
				printf("Time to transmit packet %lld\n", endmactime - mactime);
				printf("Time to retransmit small packet %lld\n", packet_retransmission_time_us(state->prev_badblocks, rate));
				printf("%lld SIM-PACKET %d %lld SPACE \n", state->mactime_actual + 
						packet_retransmission_time_us(state->prev_badblocks, rate), seqnum, 
						endmactime - mactime - 
						packet_retransmission_time_us(state->prev_badblocks, rate));

				state->time_saved_retransmit_less  += (endmactime - mactime) - 
					packet_retransmission_time_us(state->prev_badblocks, rate);

				if (state->prev_mactime != 0)
					state->mactime -= (endmactime - mactime) - 
						packet_retransmission_time_us(state->prev_badblocks, rate);

				// If there are enough good blocks in this retry, we are done
				if (good_contig_blocks(packet, len) >= state->prev_badblocks)
					state->prev_badblocks = -1;
			}
			/* How many bad blocks were in this retransmission, future retrans will need this {{{*/
			//printf("The previous packet had %d bad blocks in it\n", state->prev_badblocks);
			/*if(state->prev_badblocks > 0) {
				printf("in bad retransmission, altering state->prev_badblocks from %d to %d\n",
				state->prev_badblocks, state->prev_badblocks - 
				good_blocks_in_first_n_blocks(packet, len, state->prev_badblocks));
			// Currently Maranello does not repair incrementally it is a whole nack or nothing.
			state->prev_badblocks = state->prev_badblocks - 
			good_blocks_in_first_n_blocks(packet, len, state->prev_badblocks);
			} else {
			// TODO This should be the same for the orig bad tran, or orig retran
			state->prev_badblocks = num_blocks(packet, len) - good_blocks(packet, len);
			} }}}*/
		} /*}}}*/

		/* Bad transmission {{{*/
		else
		{
			printf("%lld PACKET %d %lld BAD_DATA %f %d\n", state->mactime_actual, seqnum,
					endmactime - mactime, rate, badblocks);
			printf("%lld SIM-PACKET %d %lld BAD_DATA %d\n", state->mactime_actual, seqnum, 
					endmactime - mactime, badblocks);
			print_good_blocks_iperf(packet, len, rate, state->mactime_actual);

			state->packets_bad++;
			/* How many bad blocks were in this retransmission, future retrans will need this */
			state->prev_badblocks = num_blocks(packet, len) - good_blocks(packet, len);
			state->prev_seq = seqnum;
		} /*}}}*/
	} 

	/* Compute actual time of packet arrival */
	if (state->prev_mactime_actual != 0)
		state->mactime_actual += (endmactime - state->prev_mactime_actual);
	else
		state->mactime_actual = endmactime - mactime;

	state->prev_mactime_actual = endmactime;
	state->prev_mactime = endmactime;

} /*}}}*/

uint64_t packet_transmission_time_us(int len, double bitrate) /*{{{*/
{
	//assert(len > 1470); /* because it should have the headers */

	int len_bit = len*8;
	/* converting to us and to bits cancels eachother out */
	double rate_us = bitrate;
	double time = (double)len_bit/rate_us;
	return floor(time);
} /*}}}*/

uint64_t packet_retransmission_time_us(int blocks, double bitrate_mbits_per_second) /*{{{*/
{

	assert(bitrate_mbits_per_second >=1);
	assert(blocks*BLOCKSIZE >= 4);
	assert(bitrate_mbits_per_second <=54);
	//assert(blocks <= 24);

	int len_bit = (WLANSIZE + 
			LLCSIZE + IPSIZE + UDPSIZE + (blocks*BLOCKSIZE))*8;
	/* converting to us and to bits cancels eachother out */
	double rate_bits_per_us = bitrate_mbits_per_second;
	double time_us = (double)len_bit/rate_bits_per_us;
	return floor(time_us);
} /*}}}*/

int is_header_correct(u_char* packet, int len) /*{{{*/
{
	/* Does the header look like a retransmission header should? */
	int rtaplen = radiotap_len(packet);
	packet += rtaplen;
	len -= rtaplen; 

	struct wlan_hdr* wlan = (struct wlan_hdr*)packet;
	struct llc_hdr* llc = (struct llc_hdr*)(packet + sizeof(struct wlan_hdr));
	struct ip_hdr* ip = (struct ip_hdr*)(packet + sizeof(struct wlan_hdr) + sizeof(struct llc_hdr));
	struct udp_hdr* udp = (struct udp_hdr*)(packet + sizeof(struct wlan_hdr) + sizeof(struct llc_hdr) + 
			sizeof(struct ip_hdr));

	/* WLAN */
	// fc 
	//printf("wlan->addr1 %d\n", !memcmp(wlan->addr1, DST, 6));
	//printf("wlan->addr2 %d\n", !memcmp(wlan->addr2, BSSID, 6));
	//printf("wlan->addr3 %d\n", !memcmp(wlan->addr3, SRC, 6));
	// duration
	// seq

	/* LLC */
	//printf("llc->dsap %d\n", llc->dsap == 0xAA);
	//printf("llc->ssap %d\n", llc->ssap == 0xAA);
	//printf("llc->control %d\n", llc->control == 0x03);
	//printf("llc->oui %d\n", !memcmp(llc->oui, LLC_ORG, 3));
	//printf("llc->type %d\n", !memcmp(llc->type, LLC_TYPE, 2));

	/* IP */
	//printf("ip->version_hdrlen %d\n", ip->version_hdrlen == 0x45);
	//printf("ip->dsfield %d\n", ip->dsfield == 0x00);
	//printf("ip->len %d\n", ntohs(ip->len) == IP_LEN);
	// do not check ip ID
	//ip->id == 0 &&
	//printf("ip->flags %d\n",  ip->flags == 0x40);
	//printf("ip->frag_offset %d\n", ip->frag_offset == 0x00);
	//printf("ip->ttl %d\n", ip->ttl == 64);
	//printf("ip->proto %d\n", ip->proto == 0x11);
	// do not check checksum
	//printf("ip->src %d\n", !memcmp(ip->src, IP_SRC, 4));
	//printf("ip->dst %d\n", !memcmp(ip->dst, IP_DST, 4));

	/* UDP */
	//htons(udp->srcport) == UDP_SRC && 
	//printf("udp->dstport %d\n", htons(udp->dstport) == UDP_DST);
	//printf("udp->len %d\n", htons(udp->len) == UDP_LEN);
	// do not check checksum

	return 
		/* WLAN */
		// fc 
		!memcmp(wlan->addr1, DST, 6) &&
		!memcmp(wlan->addr2, BSSID, 6) &&
		!memcmp(wlan->addr3, SRC, 6) &&
		// duration
		// seq

		/* LLC */
		llc->dsap == 0xAA &&
		llc->ssap == 0xAA  &&
		llc->control == 0x03 &&
		!memcmp(llc->oui, LLC_ORG, 3) &&
		!memcmp(llc->type, LLC_TYPE, 2) &&

		/* IP */
		ip->version_hdrlen == 0x45 &&
		ip->dsfield == 0x00 &&
		ntohs(ip->len) == IP_LEN &&
		// do not check ip ID
		//ip->id == 0 &&
		//ip->flags == 0x40 &&
		ip->frag_offset == 0x00 &&
		//ip->ttl == 64 &&
		ip->proto == 0x11 &&
		// do not check checksum
		!memcmp(ip->src, IP_SRC, 4) &&
		!memcmp(ip->dst, IP_DST, 4) &&

		/* UDP */
		//htons(udp->srcport) == UDP_SRC && 
		htons(udp->dstport) == UDP_DST &&
		htons(udp->len) == UDP_LEN; 	
	// do not check checksum

} /*}}}*/

void init_constants_from_packet(u_char* packet, int len) /*{{{*/
{

	int rtaplen = radiotap_len(packet);
	packet += rtaplen;
	len -= rtaplen; 

	struct wlan_hdr* wlan = (struct wlan_hdr*)packet;
	struct llc_hdr* llc = (struct llc_hdr*)(packet + sizeof(struct wlan_hdr));
	struct ip_hdr* ip = (struct ip_hdr*)(packet + sizeof(struct wlan_hdr) + sizeof(struct llc_hdr));
	struct udp_hdr* udp = (struct udp_hdr*)(packet + sizeof(struct wlan_hdr) + sizeof(struct llc_hdr) + 
			sizeof(struct ip_hdr));

	/* WLAN */
	// fc 
	memcpy(DST, wlan->addr1, 6);
	memcpy(BSSID, wlan->addr2, 6);
	memcpy(SRC, wlan->addr3, 6);
	// duration
	// seq

	/* LLC */
	memcpy(LLC_ORG, llc->oui, 3);
	memcpy(LLC_TYPE, llc->type, 2);

	/* IP */
	IP_LEN = ntohs(ip->len);
	memcpy(IP_SRC, ip->src, 4);
	memcpy(IP_DST, ip->dst, 4);

	/* UDP */
	UDP_SRC = htons(udp->srcport);
	UDP_DST = htons(udp->dstport);
	UDP_LEN = htons(udp->len); 	
} /*}}}*/

long double bandwidth(uint32_t bytes, uint64_t time) /*{{{*/
{
	long double us = time*0.000001;
	return (((long double)bytes)/131072.0)/us;
} /*}}}*/

void die(char* text) /*{{{*/
{
	printf("Error: %s\n", text);
	exit(2);
} /*}}}*/

void print_mac(const u_char* mac) /*{{{*/
{
	printf("%x:%x:%x:%x:%x:%x\n", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
} /*}}}*/

void print_simulator_state(struct simulator_state* state) /*{{{*/
{
	fprintf(stderr, "== Packets ====================\n");
	fprintf(stderr, "\tTotal Packets: %d\n", state->packets_total);
	fprintf(stderr, "\tPackets with no MAC timestamp: %f\n", (double)state->packets_nomactime/(double)state->packets_total);
	fprintf(stderr, "\tGood Packets: %f\n", (double)state->packets/(double)state->packets_total);
	fprintf(stderr, "\tBad Packets: %f\n", (double)state->packets_bad/(double)state->packets_total);
	fprintf(stderr, "\tGood Retran Packets: %f\n", (double)state->packets_retran/(double)state->packets_total);
	fprintf(stderr, "\tBad Retran Packets: %f\n", (double)state->packets_retran_bad/(double)state->packets_total);
	fprintf(stderr, "\tRetran Packets Follow Bad Packet: %f\n", (double)state->packets_retran_follows_bad/(double)state->packets_total);
	fprintf(stderr, "\tAvg Rate: %f\n", (double)state->rates_total/(double)state->packets_total);
	fprintf(stderr, "\tAvg Original Rate: %f\n", (double)state->orig_rates_total/(double)state->packets_orig);
	fprintf(stderr, "\tBlock Error Rate: %f\n", (double)state->error_blocks_total/(double)state->blocks_total);
	fprintf(stderr, "\n");
	fprintf(stderr, "== Simulated Packets ====================\n");
	fprintf(stderr, "\tTime saved because repair succeeded and retransmission is not needed: %lld us\n", state->time_saved_repair_worked);
	fprintf(stderr, "\tTime saved because retransmitting less: %lld us\n", state->time_saved_retransmit_less);
	fprintf(stderr, "\n");
	fprintf(stderr, "== Time ====================\n");
	fprintf(stderr, "\tAct MAC: %lld us\n", state->mactime_actual);
	fprintf(stderr, "\tSim MAC: %lld us\n", state->mactime);
	fprintf(stderr, "\n");
	fprintf(stderr, "== Throughput ===============\n");
	fprintf(stderr, "\tAct: %Lf Mbit/s\n", bandwidth(state->bytes_total, state->mactime_actual));
	fprintf(stderr, "\tSim: %Lf Mbit/s\n", bandwidth(state->bytes_total, state->mactime));
	fprintf(stderr, "\tSpeedup: %Lf\n", bandwidth(state->bytes_total, state->mactime)/bandwidth(state->bytes_total, state->mactime_actual));
	fprintf(stderr, "\n");
} /*}}}*/
